import sys, os
import itertools

import copy

import logging

from . import modelBase
from ...utils.geometry import getEx

import random

import orthopoly

import numpy as np
from scipy.spatial import cKDTree

from pyquaternion import Quaternion

from scipy.spatial.transform import Rotation as R
from scipy.optimize import root


class TUBE(modelBase):
    """
    Component name: TUBE
    Component type: model

    Author: Pablo Ibáñez-Freire and Pablo Palacios-Alonso
    Date: 19/05/2023

    Polimerization model for a tube.

    """

    def __generateRandomPositions(self,monomersPositions = [],monomersOrientations = []):

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        n=1
        while(n<=self.nMonomers):
            self.logger.debug(f"[TUBE] Trying to add monomer {n}")

            x=np.random.uniform(low=-boxX,high=boxX)
            y=np.random.uniform(low=-boxY,high=boxY)
            z=np.random.uniform(low=-boxZ,high=boxZ)

            currentMonomerPosition = np.asarray([x,y,z])

            if monomersPositions:
                minDst,minDstIndex = cKDTree(monomersPositions).query(currentMonomerPosition, 1)
            else:
                minDst = np.inf

            if minDst > 1.5*(2.0*self.monomerRadius):
                monomersPositions.append(currentMonomerPosition)
                q=Quaternion.random()
                q0,q1,q2,q3 = q
                monomersOrientations.append(np.asarray([q0,q1,q2,q3]))
                self.logger.debug(f"[TUBE] Added monomer {n}")
                n=n+1


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateLine(self):
        monomersPositions    = []
        monomersOrientations = []

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers):
            self.logger.debug(f"[TUBE] Trying to add monomer {n}")

            x=-boxX+self.monomerRadius+(n+1)*2.0*self.monomerRadius
            y=0.0
            z=0.0

            currentMonomerPosition = np.asarray([x,y,z])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(np.asarray([1.0,0.0,0.0,0.0]))

        centroid = np.mean(monomersPositions,axis=0)
        monomersPositions-=centroid

        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateHelix(self,yOffset=0.0,center=True):

        sigma = 2.0*self.monomerRadius

        monomersPositions    = []
        monomersOrientations = []

        alpha = np.sqrt(self.theta0T*self.theta0T+self.phi0T*self.phi0T)

        cp = self.theta0T/alpha;
        sp = self.phi0T/alpha;

        cp2 = cp*cp;
        sp2 = sp*sp;

        ca = np.cos(alpha);
        sa = np.sin(alpha);

        Rtrans = np.asarray([[ca*cp2+sp2    ,-sa*cp ,(1.0-ca)*sp*cp],
                             [sa*cp         ,ca     ,-sa*sp],
                             [(1.0-ca)*sp*cp,sa*sp  ,ca*sp2+cp2]])

        r_trans = R.from_matrix(Rtrans)

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers):
            self.logger.debug(f"[TUBE] Adding monomer {n}")

            if n==0:
                x=0.0
                y=-boxY+sigma+yOffset
                z=(self.theta0T/(self.theta0T*self.theta0T+self.phi0T*self.phi0T)-1.0)

                Rinit = np.asarray([[ca*cp ,-sa ,-ca*sp],
                                    [sp    ,0.0 , cp],
                                    [-sa*cp,-ca , sa*sp]])

                r_init = R.from_matrix(Rinit)

                q1,q2,q3,q0 = r_init.as_quat()

            else:
                q0,q1,q2,q3 = monomersOrientations[-1]
                r_prev    = R.from_quat([q1,q2,q3,q0])
                r_current = r_prev*r_trans

                q1,q2,q3,q0 = r_current.as_quat()

                x,y,z = monomersPositions[-1]+0.5*sigma*(getEx(monomersOrientations[-1])+getEx([q0,q1,q2,q3]))


            currentMonomerPosition = np.asarray([x,y,z])
            currentMonomerOrientation = np.asarray([q0,q1,q2,q3])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(currentMonomerOrientation)

        if(center):
            centroid = np.mean(monomersPositions,axis=0)
            monomersPositions-=centroid

        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __getMonomerConnections(self,position,orientation):
        connectionStart = position-self.monomerRadius*getEx(orientation)
        connectionEnd   = position+self.monomerRadius*getEx(orientation)

        return connectionStart,connectionEnd

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"init",
                                                "nMonomers","box",
                                                "monomerRadius","patchRadius",
                                                "pitch","monomersPerTurn",
                                                "epsilon_mm",
                                                "EbT","rcT",
                                                "KbT","KaT","KdT",
                                                "stiffnessFactorT",
                                                "EbL","rcL",
                                                "KbL","KaL","KdL",
                                                "theta0L",
                                                "phi0L",
                                                "stiffnessFactorL"},
                         requiredParameters  = {"nMonomers","box",
                                                "pitch","monomersPerTurn",
                                                "epsilon_mm",
                                                "EbT","rcT",
                                                "KbT","KaT","KdT",
                                                "EbL","rcL",
                                                "KbL","KaL","KdL",
                                                "theta0L"},
                         definedSelections   = {"particleId"},
                         **params)

        self.logger = logging.getLogger("VLMP")

        ############################################################
        ####################  Model Parameters  ####################
        ############################################################

        self.kT = self.getUnits().getConstant("kT")

        self.init      = params.get("init","random")

        self.nMonomers      = params["nMonomers"]

        self.box = params["box"]

        self.monomerRadius = params.get("monomerRadius",0.5)
        self.patchRadius   = params.get("patchRadius",0.1)

        self.pitch           = params["pitch"]
        self.monomersPerTurn = params["monomersPerTurn"]

        self.epsilon_mm = params["epsilon_mm"]

        #Transversal

        self.EbT = params["EbT"]
        self.rcT = params["rcT"]

        self.KbT = params["KbT"]
        self.KaT = params["KaT"]
        self.KdT = params["KdT"]

        stiffnessFactor = params.get("stiffnessFactorT",1.0)
        if stiffnessFactor != 1.0:
            self.logger.info(f"[TUBE] Applying stiffness factor (T) {stiffnessFactor}")
            self.KaT*=stiffnessFactor
            self.KbT*=stiffnessFactor

            self.logger.info(f"[TUBE] Kt theta after stiffness factor {self.KaT}")
            self.logger.info(f"[TUBE] Kt phi after stiffness factor {self.KdT}")

        def equationSystem(vars, nc, pitch, sigma):
            phi0, theta0 = vars

            a = pitch/(np.pi*2)
            b = nc*sigma/(np.pi*2)

            R = np.sqrt(b**2-a**2)

            alpha = np.sqrt(phi0**2+theta0**2)

            eq1 = phi0/alpha - pitch/(sigma*nc)
            eq2 = np.cos(alpha)*(theta0**2/alpha**2)+phi0**2/alpha**2-(a**2+np.cos(2*np.pi/nc)*(R**2))/(a**2+R**2)

            return [eq1, eq2]

        self.phi0T,self.theta0T = root(equationSystem,
                                       [np.pi/2,np.pi/2],
                                       args=(self.monomersPerTurn,self.pitch,self.monomerRadius*2.0),tol=1e-12).x

        self.logger.info(f"[TUBE] Computing theta0T {self.theta0T}")
        self.logger.info(f"[TUBE] Computing phi0T {self.phi0T}")

        #Longitudinal

        self.EbL = params["EbL"]
        self.rcL = params["rcL"]

        self.KbL = params["KbL"]
        self.KaL = params["KaL"]
        self.KdL = params["KdL"]

        stiffnessFactor = params.get("stiffnessFactorL",1.0)
        if stiffnessFactor != 1.0:
            self.logger.info(f"[TUBE] Applying stiffness factor (L) {stiffnessFactor}")
            self.KaL*=stiffnessFactor
            self.KbL*=stiffnessFactor

            self.logger.info(f"[TUBE] Kl theta after stiffness factor {self.KaL}")
            self.logger.info(f"[TUBE] Kl phi after stiffness factor {self.KdL}")

        self.theta0L = params["theta0L"]
        self.phi0L   = params.get("phi0L",0.0)

        alpha = np.sqrt(self.theta0T**2+self.phi0T**2)

        cp = self.theta0T/alpha;
        sp = self.phi0T/alpha;

        ca = np.cos(alpha);
        sa = np.sin(alpha);

        rot = np.asarray([[ca*cp ,-sa ,-ca*sp],
                          [sp    ,0.0 , cp],
                          [-sa*cp,-ca , sa*sp]])

        vector = np.asarray([0,1,0])*self.monomerRadius*2.0


        self.Lplus  = np.dot(rot.T,vector)*self.monomerRadius
        self.Lminus = -copy.deepcopy(self.Lplus)

        ##############################################

        self.logger.info(f"[TUBE] Generating TUBE model with {self.nMonomers} monomers")

        if   self.init == "random":
            self.monomersPositions,self.monomersOrientations = self.__generateRandomPositions()
        elif self.init == "line":
            self.monomersPositions,self.monomersOrientations = self.__generateLine()
        elif self.init == "helix":
            self.monomersPositions,self.monomersOrientations = self.__generateHelix()
        elif self.init == "tube":
            nTotalMonomers = self.nMonomers #Backup

            monomersPerTube = [0 for i in range(self.pitch)]
            n=0
            i=0
            while n < self.nMonomers:
                if i >= self.pitch:
                    i=0
                monomersPerTube[i]+=1
                n+=1
                i+=1

            self.monomersPositions = np.zeros((0,3))
            self.monomersOrientations = np.zeros((0,4))
            for i in range(self.pitch):
                self.nMonomers = monomersPerTube[i]
                mPos,mOri = self.__generateHelix(yOffset=i*2.0*self.monomerRadius,center=False)
                self.monomersPositions    = np.vstack((self.monomersPositions,mPos))
                self.monomersOrientations = np.vstack((self.monomersOrientations,mOri))

            centroid = np.mean(self.monomersPositions,axis=0)
            self.monomersPositions-=centroid

            self.nMonomers = nTotalMonomers #Restore
        elif self.init == "nucleus":
            nTotalMonomers = self.nMonomers #Backup
            nucleusMonomers = self.monomersPerTurn*2*self.pitch

            if nucleusMonomers > self.nMonomers:
                logging.error(f"[TUBE] Nucleus monomers {nucleusMonomers} larger than total monomers {self.nMonomers}")
                raise Exception("Nucleus monomers larger than total monomers")

            monomersPerTube = [0 for i in range(self.pitch)]
            n=0
            i=0
            while n < nucleusMonomers:
                if i >= self.pitch:
                    i=0
                monomersPerTube[i]+=1
                n+=1
                i+=1

            self.monomersPositions = np.zeros((0,3))
            self.monomersOrientations = np.zeros((0,4))
            for i in range(self.pitch):
                self.nMonomers = monomersPerTube[i]
                mPos,mOri = self.__generateHelix(yOffset=i*2.0*self.monomerRadius,center=False)
                self.monomersPositions    = np.vstack((self.monomersPositions,mPos))
                self.monomersOrientations = np.vstack((self.monomersOrientations,mOri))

            centroid = np.mean(self.monomersPositions,axis=0)
            self.monomersPositions-=centroid

            self.nMonomers = nTotalMonomers - nucleusMonomers

            self.monomersPositions,self.monomersOrientations = self.__generateRandomPositions(self.monomersPositions.tolist(),
                                                                                              self.monomersOrientations.tolist())

            self.nMonomers = nTotalMonomers #Restore


        else:
            self.logger.error(f"[TUBE] Init mode {self.init} is not avaible")
            raise Exception("Init mode not available")

        ##############################################

        types = self.getTypes()

        types.addType(name="M",radius=self.monomerRadius)

        ##############################################

        state = {}
        state["labels"]=["id","position","direction"]
        state["data"]  =[]

        for i,[p,q] in enumerate(zip(self.monomersPositions,self.monomersOrientations)):
            p=np.around(p,2)
            q=np.around(q,2)
            state["data"].append([int(i),list(p),list(q)])

        ##############################################

        structure = {}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        for i in range(self.nMonomers):
            structure["data"].append([int(i),"M",int(i)])

        ##############################################

        forceField = {}

        #Verlet list
        forceField["verletListParticles"]={}
        forceField["verletListParticles"]["type"]       =  ["VerletConditionalListSet", "all"]
        forceField["verletListParticles"]["parameters"] =  {"cutOffVerletFactor":1.5}

        #WCA
        forceField["MonomerMonomer"]={}
        forceField["MonomerMonomer"]["type"]       =  ["NonBonded", "GeneralLennardJonesType2"]
        forceField["MonomerMonomer"]["parameters"] =  {"cutOffFactor":2.5,"condition":"all"}
        forceField["MonomerMonomer"]["labels"]     =  ["name_i", "name_j", "epsilon", "sigma"]
        forceField["MonomerMonomer"]["data"]       =  [["M", "M", self.epsilon_mm, 2.0*self.monomerRadius]]

        ############Patches

        forceField["tube"]={}
        forceField["tube"]["type"]       = ["PatchyParticles","DynamicallyBondedPatchyParticles"]

        forceField["tube"]["patchesState"]={}
        forceField["tube"]["patchesState"]["labels"]=["id","position"]
        forceField["tube"]["patchesState"]["data"]  =[]

        for i in range(self.nMonomers):

            #Transversal
            index=i*4
            p = [-self.monomerRadius,0.0,0.0]
            forceField["tube"]["patchesState"]["data"].append([int(index  ),list(p)])
            p = [ self.monomerRadius,0.0,0.0]
            forceField["tube"]["patchesState"]["data"].append([int(index+1),list(p)])

            #Longitudinal
            index=i*4
            p = copy.deepcopy(self.Lplus)
            forceField["tube"]["patchesState"]["data"].append([int(index+2),list(p)])
            p = copy.deepcopy(self.Lminus)
            forceField["tube"]["patchesState"]["data"].append([int(index+3),list(p)])

        forceField["tube"]["patchesGlobal"]={}
        forceField["tube"]["patchesGlobal"]["fundamental"] = {}
        forceField["tube"]["patchesGlobal"]["fundamental"]["type"]      = ["Fundamental","DynamicallyBondedPatchyParticles"]
        forceField["tube"]["patchesGlobal"]["fundamental"]["parameters"]={"energyThreshold":0.0}

        forceField["tube"]["patchesGlobal"]["types"]  = {}
        forceField["tube"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
        forceField["tube"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
        forceField["tube"]["patchesGlobal"]["types"]["data"]   = [["St",0.0,self.patchRadius,0.0],
                                                                  ["Et",0.0,self.patchRadius,0.0],
                                                                  ["Sl",0.0,self.patchRadius,0.0],
                                                                  ["El",0.0,self.patchRadius,0.0]]

        forceField["tube"]["patchesTopology"]={}

        forceField["tube"]["patchesTopology"]["structure"]={}
        forceField["tube"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
        forceField["tube"]["patchesTopology"]["structure"]["data"]   = []

        for i in range(self.nMonomers):
            index=i*4
            forceField["tube"]["patchesTopology"]["structure"]["data"].append([index  ,"Et",i])
            forceField["tube"]["patchesTopology"]["structure"]["data"].append([index+1,"St",i])
            forceField["tube"]["patchesTopology"]["structure"]["data"].append([index+2,"El",i])
            forceField["tube"]["patchesTopology"]["structure"]["data"].append([index+3,"Sl",i])

        forceField["tube"]["patchesTopology"]["forceField"] = {}

        forceField["tube"]["patchesTopology"]["forceField"]["groups"] = {}
        forceField["tube"]["patchesTopology"]["forceField"]["groups"]["type"] = ["Groups","GroupsList"]
        forceField["tube"]["patchesTopology"]["forceField"]["groups"]["parameters"] = {}
        forceField["tube"]["patchesTopology"]["forceField"]["groups"]["labels"] = ["name", "type", "selection"]
        forceField["tube"]["patchesTopology"]["forceField"]["groups"]["data"]   = [
            ["trans","Types",["St","Et"]],
            ["long","Types",["Sl","El"]]
        ]


        #Verlet list
        forceField["tube"]["patchesTopology"]["forceField"]["verletList"]={}
        forceField["tube"]["patchesTopology"]["forceField"]["verletList"]["type"]       =  ["VerletConditionalListSet", "interDifferentType"]
        forceField["tube"]["patchesTopology"]["forceField"]["verletList"]["parameters"] =  {"cutOffVerletFactor":3.0}

        forceField["tube"]["patchesTopology"]["forceField"]["transversal"]={}
        forceField["tube"]["patchesTopology"]["forceField"]["transversal"]["type"]       =  ["NonBondedPatches", "Helix"]
        forceField["tube"]["patchesTopology"]["forceField"]["transversal"]["parameters"] =  {"condition":"inter",
                                                                                             "startType":"St",
                                                                                             "endType":"Et",
                                                                                             "group":"trans"}
        forceField["tube"]["patchesTopology"]["forceField"]["transversal"]["labels"]     =  ["name_i", "name_j", "Eb", "Kb", "Ka", "Kd", "rc", "theta0", "phi0"]
        forceField["tube"]["patchesTopology"]["forceField"]["transversal"]["data"]       =  [["St", "Et",
                                                                                              self.EbT,
                                                                                              self.KbT, self.KaT, self.KdT,
                                                                                              self.rcT,
                                                                                              self.theta0T, self.phi0T]]
        forceField["tube"]["patchesTopology"]["forceField"]["longitudinal"]={}
        forceField["tube"]["patchesTopology"]["forceField"]["longitudinal"]["type"]       =  ["NonBondedPatches", "Helix"]
        forceField["tube"]["patchesTopology"]["forceField"]["longitudinal"]["parameters"] =  {"condition":"inter",
                                                                                              "startType":"Sl",
                                                                                              "endType":"El",
                                                                                              "group":"long"}
        forceField["tube"]["patchesTopology"]["forceField"]["longitudinal"]["labels"]     =  ["name_i", "name_j", "Eb", "Kb", "Ka", "Kd", "rc", "theta0", "phi0"]
        forceField["tube"]["patchesTopology"]["forceField"]["longitudinal"]["data"]       =  [["Sl", "El",
                                                                                              self.EbL,
                                                                                              self.KbL, self.KaL, self.KdL,
                                                                                              self.rcL,
                                                                                              self.theta0L, self.phi0L]]

        ##############################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)

    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel



import sys, os

import logging

from . import unitsBase

class KcalMol_A(unitsBase):

    """
    Component name: KcalMol_A
    Component type: units

    Author: Pablo Ibáñez-Freire
    Date: 13/03/2023

    (Kcal/mol)/A

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = set(),
                         requiredParameters  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setUnitsName("KcalMol_A")

        self.addConstant("kT",0.593)
        self.addConstant("KBOLTZ",1.987191E-03)
        self.addConstant("ELECOEF",332.0716)



from VLMP.components.models import modelBase

from VLMP.components.models.ICOSPHERE import ICOSPHERE

import numpy as np

class CORONAVIRUS_MESH(modelBase):
    """
    Component name: CORONAVIRUS_MESH
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 26/10/2023
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"particleName",
                                                "particleMass","particleRadius","particleCharge",
                                                "position",
                                                "resolution",
                                                "radius",
                                                "Kb",
                                                "Kd",
                                                "steric",
                                                "surface",
                                                "surfacePosition",
                                                "surfaceEpsilon_kT"},
                         requiredParameters  = {"particleName"},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################

        paramsICO = params.copy()

        # Remove parameters that are not needed for ICOSPHERE
        paramsICO.pop("surface")
        paramsICO.pop("surfacePosition")
        paramsICO.pop("surfaceEpsilon_kT")

        self.SPHERE = ICOSPHERE(name = name + "_sphere",
                                **paramsICO)

        units      = self.getUnits()

        types      = self.SPHERE.getTypes()
        state      = self.SPHERE.getState()
        structure  = self.SPHERE.getStructure()
        forceField = self.SPHERE.getForceField()

        surface = params.get("surface",False)
        if surface:
            surfaceEpsilon  = params["surfaceEpsilon_kT"]*units.getConstant("kT")
            surfacePosition = params["surfacePosition"]

        ############################################################

        if surface:

            forceField["surface"] = {}
            forceField["surface"]["type"]       = ["Surface", "SurfaceGeneralLennardJonesType2"]
            forceField["surface"]["parameters"] = {"surfacePosition":surfacePosition}
            forceField["surface"]["labels"] = ["name","epsilon","sigma"]
            forceField["surface"]["data"]   = []

            for tn,tinfo in types.getTypes().items():
                forceField["surface"]["data"].append([tinfo["name"],surfaceEpsilon,tinfo["radius"]])

        ############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = self.SPHERE.processSelection(**params)

        return sel




import sys, os
import itertools

import logging

from . import modelBase
from ...utils.geometry import getEx
from ...utils.geometry import boxChecker,platesChecker

from ...utils.input import getSubParameters

import random

import orthopoly

import numpy as np
from scipy.spatial import cKDTree

from pyquaternion import Quaternion

from scipy.spatial.transform import Rotation as R
from scipy.optimize import root


class HELIX(modelBase):
    """
    Component name: HELIX
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 25/04/2023

    Polimerization model for a helix.

    """

    def __computeKdst(self,var,Eb,rc):
        init = 1.0
        Kdst = root(lambda K: var - (self.kT/Eb)*((rc*rc)/(np.pi*np.pi))*((1.0-np.exp(-2.0*K))/K),init)
        varTheo = (self.kT/Eb)*((rc*rc)/(np.pi*np.pi))*((1.0-np.exp(-2.0*Kdst["x"][0]))/Kdst["x"][0])
        if np.abs(varTheo-var) > 1e-6:
            self.logger.error(f"[HELIX] Error computing Kdst")
            raise Exception("Error computing Kdst")
        return Kdst["x"][0]

    def __computeKangle(self,var,Eb):
        init = (1.0-0.5*var*(Eb/self.kT))/2.0
        Kangle = root(lambda K: var - (self.kT/Eb)*((1.0-np.exp(-2.0*K))/K),init)
        varTheo = (self.kT/Eb)*((1.0-np.exp(-2.0*Kangle["x"][0]))/Kangle["x"][0])
        if np.abs(varTheo-var) > 1e-6:
            self.logger.error(f"[HELIX] Error computing Kangle")
            raise Exception("Error computing Kangle")
        return Kangle["x"][0]


    def __generateRandomPositions(self):
        monomersPositions    = []
        monomersOrientations = []

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        n=1
        while(n<=self.nMonomers):
            self.logger.debug(f"[HELIX] Trying to add monomer {n}")

            #We take into account the monomer radius for avoiding problems with PBC
            x=np.random.uniform(low=-boxX + self.monomerRadius, high=boxX - self.monomerRadius)
            y=np.random.uniform(low=-boxY + self.monomerRadius, high=boxY - self.monomerRadius)
            z=np.random.uniform(low=-boxZ + self.monomerRadius, high=boxZ - self.monomerRadius)

            currentMonomerPosition = np.asarray([x,y,z])

            if not self.checker.check([x,y,z]):
                continue

            if monomersPositions:
                minDst,minDstIndex = cKDTree(monomersPositions).query(currentMonomerPosition, 1)
            else:
                minDst = np.inf

            if minDst > 1.5*(2.0*self.monomerRadius):
                monomersPositions.append(currentMonomerPosition)
                q=Quaternion.random()
                q0,q1,q2,q3 = q
                monomersOrientations.append(np.asarray([q0,q1,q2,q3]))
                self.logger.debug(f"[HELIX] Added monomer {n}")
                n=n+1


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateLine(self):
        monomersPositions    = []
        monomersOrientations = []

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers):
            self.logger.debug(f"[HELIX] Trying to add monomer {n}")

            x=-boxX+self.monomerRadius+(n+1)*2.2*self.monomerRadius
            y=0.0
            z=0.0

            currentMonomerPosition = np.asarray([x,y,z])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(np.asarray([1.0,0.0,0.0,0.0]))

        centroid = np.mean(monomersPositions,axis=0)
        monomersPositions-=centroid

        for p in monomersPositions:
            if not self.checker.check(p):
                self.logger.error(f"[HELIX] Box too small")
                raise Exception("Box too small")


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __generateHelix(self):

        sigma = 2.0*self.monomerRadius

        monomersPositions    = []
        monomersOrientations = []

        alpha = np.sqrt(self.theta0*self.theta0+self.phi0*self.phi0)

        cp = self.theta0/alpha;
        sp = self.phi0/alpha;

        cp2 = cp*cp;
        sp2 = sp*sp;

        ca = np.cos(alpha);
        sa = np.sin(alpha);

        Rtrans = np.asarray([[ca*cp2+sp2    ,-sa*cp ,(1.0-ca)*sp*cp],
                             [sa*cp         ,ca     ,-sa*sp],
                             [(1.0-ca)*sp*cp,sa*sp  ,ca*sp2+cp2]])

        r_trans = R.from_matrix(Rtrans)

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers):
            self.logger.debug(f"[HELIX] Adding monomer {n}")

            if n==0:
                x=0.0
                y=-boxY+sigma
                z=(self.theta0/(self.theta0*self.theta0+self.phi0*self.phi0)-1.0)

                Rinit = np.asarray([[ca*cp ,-sa ,-ca*sp],
                                    [sp    ,0.0 , cp],
                                    [-sa*cp,-ca , sa*sp]])

                r_init = R.from_matrix(Rinit)

                q1,q2,q3,q0 = r_init.as_quat()

            else:

                q0,q1,q2,q3 = monomersOrientations[-1]
                r_prev    = R.from_quat([q1,q2,q3,q0])
                r_current = r_prev*r_trans

                x,y,z = monomersPositions[-1]+0.5*sigma*(getEx(monomersOrientations[-1])+getEx([q0,q1,q2,q3]))

                q1,q2,q3,q0 = r_current.as_quat()


            currentMonomerPosition = np.asarray([x,y,z])
            currentMonomerOrientation = np.asarray([q0,q1,q2,q3])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(currentMonomerOrientation)

        centroid = np.mean(monomersPositions,axis=0)
        monomersPositions-=centroid

        for p in monomersPositions:
            if not self.checker.check(p):
                self.logger.error(f"[HELIX] Box too small")
                raise Exception("Box too small")


        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __getMonomerConnections(self,position,orientation):
        connectionStart = position-self.monomerRadius*getEx(orientation)
        connectionEnd   = position+self.monomerRadius*getEx(orientation)

        return connectionStart,connectionEnd

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"mode","init",
                                                "nMonomers",
                                                "monomerRadius","patchRadius",
                                                "epsilon_mm",
                                                "Eb","rc",
                                                "theta0","phi0",
                                                "varDst","varTheta","varPhi",
                                                "Kb","Ka","Kd",
                                                "energyThreshold",
                                                "stiffnessFactor",
                                                "variant"},
                         requiredParameters  = {"nMonomers",
                                                "epsilon_mm",
                                                "Eb","rc",
                                                "theta0","phi0"},
                         definedSelections   = {"particleId"},
                         **params)

        self.logger = logging.getLogger("VLMP")

        ############################################################
        ####################  Model Parameters  ####################
        ############################################################

        self.kT = self.getUnits().getConstant("kT")

        self.init                 = params.get("init","random")
        self.mode,self.modeParams = getSubParameters("mode",params)

        if self.mode not in ["bulk","surface"]:
            self.logger.error(f"[HELIX] Mode {self.mode} not supported")
            raise Exception(f"Mode not supported")

        self.nMonomers      = params["nMonomers"]

        self.box = self.getEnsemble().getEnsembleComponent("box")

        self.monomerRadius = params.get("monomerRadius",0.5)
        self.patchRadius   = params.get("patchRadius",0.1)

        self.epsilon_mm = params["epsilon_mm"]

        self.Eb = params["Eb"]
        self.rc = params["rc"]

        self.energyThreshold = params.get("energyThreshold",-1.0)

        stiffnessFactor = params.get("stiffnessFactor",1.0)

        self.theta0 = params["theta0"]
        self.phi0   = params["phi0"]

        if self.mode=="surface":
            #Check parameters Es,beta0,El,Sl are present in params
            requiredSurfaceParams = ["Es","beta0","El","Sl","plateTop","plateBottom"]

            for sPar in requiredSurfaceParams:
                if sPar not in self.modeParams:
                    self.logger.error(f"[HELIX] Requiered parameter {sPar} for surface mode not found."
                                      f"Required parameters are {requiredSurfaceParams}")
                    raise Exception(f"Required parameter not given")

            self.Es     = self.modeParams["Es"]

            self.beta0  = self.modeParams["beta0"]

            self.El     = self.modeParams["El"]
            self.Sl     = self.modeParams["Sl"]

            self.plateTop    = self.modeParams["plateTop"]
            self.plateBottom = self.modeParams["plateBottom"]

            self.checker = platesChecker(self.box,self.plateTop,self.plateBottom,2.0*self.monomerRadius)
        else:
            self.checker = boxChecker(self.box)

        if "varDst" in params:
            varDst   = params["varDst"]
            self.Kb = self.__computeKdst(varDst,abs(self.Eb),self.rc)
            self.logger.info(f"[HELIX] K bond computed {self.Kb} for variance {varDst}")
        else:
            self.Kb = params["Kb"]
            self.logger.info(f"[HELIX] K bond {self.Kb}")

        if "varTheta" in params:
            varTheta = params["varTheta"]
            self.Ka = self.__computeKangle(varTheta,abs(self.Eb))
            self.logger.info(f"[HELIX] K theta computed {self.Ka} for variance {varTheta}")
        else:
            self.Ka = params["Ka"]
            self.logger.info(f"[HELIX] K theta {self.Ka}")

        if "varPhi" in params:
            varPhi   = params["varPhi"]
            self.Kd = self.__computeKangle(varPhi,abs(self.Eb))
            self.logger.info(f"[HELIX] K phi computed {self.Kd} for variance {varPhi}")
        else:
            self.Kd = params["Kd"]
            self.logger.info(f"[HELIX] K phi {self.Kd}")

        if stiffnessFactor != 1.0:
            self.logger.info(f"[HELIX] Applying stiffness factor {stiffnessFactor}")
            self.Ka*=stiffnessFactor
            self.Kd*=stiffnessFactor

            self.logger.info(f"[HELIX] K theta after stiffness factor {self.Ka}")
            self.logger.info(f"[HELIX] K phi after stiffness factor {self.Kd}")

        if self.Kb < 0 or self.Ka < 0 or self.Kd < 0:
            self.logger.error(f"[HELIX] Negative spring constant: Kb {self.Kb} Ka {self.Ka} Kd {self.Kd}")
            raise Exception("Negative spring constant")

        ##############################################

        self.variantName, self.variantParams = getSubParameters("variant",params)

        if self.variantName == "twoStates":
            self.logger.info(f"[HELIX] Using two states variant")

            variantAvailableParameters = {"Eb2","rc2",
                                          "theta02","phi02",
                                          "varDst2","varTheta2","varPhi2",
                                          "Kb2","Ka2","Kd2",
                                          "prob_1_to_2","prob_2_to_1"}

            variantRequiredParameters  = {"Eb2","rc2",
                                          "theta02","phi02",
                                          "prob_1_to_2","prob_2_to_1"}

            for key in self.variantParams.keys():
                if key not in variantAvailableParameters:
                    self.logger.error(f"[HELIX] Variant parameter {key} is not available for variant {self.variantName}")
                    raise Exception("Not correct variant parameter")

            for key in variantRequiredParameters:
                if key not in self.variantParams.keys():
                    self.logger.error(f"[HELIX] Variant parameter {key} is required for variant {self.variantName}")
                    raise Exception("Not correct variant parameter")

            self.Eb2 = self.variantParams["Eb2"]
            self.rc2 = self.variantParams["rc2"]

            self.theta02 = self.variantParams["theta02"]
            self.phi02   = self.variantParams["phi02"]

            if "varDst2" in self.variantParams:
                varDst2   = self.variantParams["varDst2"]
                self.Kb2 = self.__computeKdst(varDst2,abs(self.Eb2),self.rc2)
                self.logger.info(f"[HELIX] K bond 2 computed {self.Kb2} for variance {varDst2}")
            else:
                self.Kb2 = self.variantParams["Kb2"]
                self.logger.info(f"[HELIX] K bond 2 {self.Kb2}")

            if "varTheta2" in self.variantParams:
                varTheta2 = self.variantParams["varTheta2"]
                self.Ka2 = self.__computeKangle(varTheta2,abs(self.Eb2))
                self.logger.info(f"[HELIX] K theta 2 computed {self.Ka2} for variance {varTheta2}")
            else:
                self.Ka2 = self.variantParams["Ka2"]
                self.logger.info(f"[HELIX] K theta 2 {self.Ka2}")

            if "varPhi2" in self.variantParams:
                varPhi2   = self.variantParams["varPhi2"]
                self.Kd2 = self.__computeKangle(varPhi2,abs(self.Eb2))
                self.logger.info(f"[HELIX] K phi 2 computed {self.Kd2} for variance {varPhi2}")
            else:
                self.Kd2 = self.variantParams["Kd2"]
                self.logger.info(f"[HELIX] K phi 2 {self.Kd2}")

            if stiffnessFactor != 1.0:
                self.logger.info(f"[HELIX] Applying stiffness factor {stiffnessFactor}")

                self.Ka2*=stiffnessFactor
                self.Kd2*=stiffnessFactor

                self.logger.info(f"[HELIX] K2 theta after stiffness factor {self.Ka2}")
                self.logger.info(f"[HELIX] K2 phi after stiffness factor {self.Kd2}")

            if self.Kb2 < 0 or self.Ka2 < 0 or self.Kd2 < 0:
                self.logger.error(f"[HELIX] Negative spring constant: Kb2 {self.Kb2} Ka2 {self.Ka2} Kd2 {self.Kd2}")
                raise Exception("Negative spring constant")

            self.prob_0_to_1 = self.variantParams["prob_1_to_2"] #Note the different naming convention
            self.prob_1_to_0 = self.variantParams["prob_2_to_1"]

        ##############################################

        self.logger.info(f"[HELIX] Generating HELIX model with {self.nMonomers} monomers")

        if   self.init == "random":
            self.monomersPositions,self.monomersOrientations = self.__generateRandomPositions()
        elif self.init == "line":
            self.monomersPositions,self.monomersOrientations = self.__generateLine()
        elif self.init == "helix":
            self.monomersPositions,self.monomersOrientations = self.__generateHelix()
        else:
            self.logger.error(f"[HELIX] Init mode {self.init} is not avaible")
            raise Exception("Init mode not available")

        ##############################################

        types = self.getTypes()

        types.addType(name="M",radius=self.monomerRadius)

        ##############################################

        state = {}
        state["labels"]=["id","position","direction"]
        state["data"]  =[]

        for i,[p,q] in enumerate(zip(self.monomersPositions,self.monomersOrientations)):
            p=np.around(p,2)
            q=np.around(q,2)
            state["data"].append([int(i),list(p),list(q)])

        ##############################################

        structure = {}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        for i in range(self.nMonomers):
            structure["data"].append([int(i),"M",int(i)])

        ##############################################

        forceField = {}

        #Verlet list
        forceField["verletListParticles"]={}
        forceField["verletListParticles"]["type"]       =  ["VerletConditionalListSet", "all"]
        forceField["verletListParticles"]["parameters"] =  {"cutOffVerletFactor":1.5}

        #WCA
        forceField["MonomerMonomer"]={}
        forceField["MonomerMonomer"]["type"]       =  ["NonBonded", "GeneralLennardJonesType2"]
        forceField["MonomerMonomer"]["parameters"] =  {"cutOffFactor":2.5,"condition":"all"}
        forceField["MonomerMonomer"]["labels"]     =  ["name_i", "name_j", "epsilon", "sigma"]
        forceField["MonomerMonomer"]["data"]       =  [["M", "M", self.epsilon_mm, 2.0*self.monomerRadius]]

        ############Patches

        #Helix
        forceField["helix"]={}
        forceField["helix"]["type"]       = ["PatchyParticles","DynamicallyBondedPatchyParticles"]

        forceField["helix"]["patchesState"]={}
        forceField["helix"]["patchesState"]["labels"]=["id","position"]
        forceField["helix"]["patchesState"]["data"]  =[]

        for i in range(self.nMonomers):
            index=i*2
            p = [-self.monomerRadius,0.0,0.0]
            forceField["helix"]["patchesState"]["data"].append([int(index  ),list(p)])
            p = [ self.monomerRadius,0.0,0.0]
            forceField["helix"]["patchesState"]["data"].append([int(index+1),list(p)])

        forceField["helix"]["patchesGlobal"]={}
        forceField["helix"]["patchesGlobal"]["fundamental"] = {}
        forceField["helix"]["patchesGlobal"]["fundamental"]["type"]       = ["Fundamental","DynamicallyBondedPatchyParticles"]
        forceField["helix"]["patchesGlobal"]["fundamental"]["parameters"] = {"energyThreshold":self.energyThreshold}

        forceField["helix"]["patchesGlobal"]["types"]  = {}
        forceField["helix"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
        forceField["helix"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
        forceField["helix"]["patchesGlobal"]["types"]["data"]   = [["S",0.0,self.patchRadius,0.0],
                                                                   ["E",0.0,self.patchRadius,0.0]]

        forceField["helix"]["patchesTopology"]={}

        forceField["helix"]["patchesTopology"]["structure"]={}
        forceField["helix"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
        forceField["helix"]["patchesTopology"]["structure"]["data"]   = []

        for i in range(self.nMonomers):
            index=i*2
            forceField["helix"]["patchesTopology"]["structure"]["data"].append([index  ,"E",i])
            forceField["helix"]["patchesTopology"]["structure"]["data"].append([index+1,"S",i])

        forceField["helix"]["patchesTopology"]["forceField"] = {}

        #Verlet list
        forceField["helix"]["patchesTopology"]["forceField"]["verletList"]={}
        forceField["helix"]["patchesTopology"]["forceField"]["verletList"]["type"]       =  ["VerletConditionalListSet", "interDifferentType"]
        forceField["helix"]["patchesTopology"]["forceField"]["verletList"]["parameters"] =  {"cutOffVerletFactor":3.0}

        forceField["helix"]["patchesTopology"]["forceField"]["helix"]={}
        forceField["helix"]["patchesTopology"]["forceField"]["helix"]["type"]       =  ["NonBondedPatches", "Helix"]
        forceField["helix"]["patchesTopology"]["forceField"]["helix"]["parameters"] =  {"condition":"inter"}
        forceField["helix"]["patchesTopology"]["forceField"]["helix"]["labels"]     =  ["name_i", "name_j", "Eb", "Kb", "Ka", "Kd", "rc", "theta0", "phi0"]
        forceField["helix"]["patchesTopology"]["forceField"]["helix"]["data"]       =  [["S", "E",
                                                                                         self.Eb,
                                                                                         self.Kb, self.Ka, self.Kd,
                                                                                         self.rc,
                                                                                         self.theta0, self.phi0]]

        if self.variantName == "twoStates":
            forceField["helix"]["patchesTopology"]["forceField"]["helix"]["type"]   =  ["NonBondedPatches", "Helix2States"]
            forceField["helix"]["patchesTopology"]["forceField"]["helix"]["labels"] = ["name_i", "name_j",
                                                                                       "Eb0", "Kb0", "Ka0", "Kd0", "rc0", "theta0_0", "phi0_0",
                                                                                       "Eb1", "Kb1", "Ka1", "Kd1", "rc1", "theta0_1", "phi0_1",
                                                                                       "prob_0_to_1","prob_1_to_0"]
            forceField["helix"]["patchesTopology"]["forceField"]["helix"]["data"][-1].extend([self.Eb2,
                                                                                              self.Kb2, self.Ka2, self.Kd2,
                                                                                              self.rc2,
                                                                                              self.theta02, self.phi02,
                                                                                              self.prob_0_to_1,self.prob_1_to_0])

        ############Patches

        #Surface
        if(self.mode == "surface"):

            forceField["plateUp"]={}
            forceField["plateUp"]["type"]           = ["Surface", "SurfaceWCAType2"]
            forceField["plateUp"]["parameters"]     = {"surfacePosition": self.plateTop}
            forceField["plateUp"]["labels"]         = ["name", "epsilon", "sigma"]
            forceField["plateUp"]["data"]           = [["M",1.0,2.0*self.monomerRadius]]

            forceField["plateBottom"]={}
            forceField["plateBottom"]["type"]       = ["Surface", "ParabolaSurface"]
            forceField["plateBottom"]["parameters"] = {"surfacePosition":self.plateBottom+self.monomerRadius}
            forceField["plateBottom"]["labels"]     = ["name", "epsilon"]
            forceField["plateBottom"]["data"]       = [["M",self.Es]]

            #Patches

            forceField["surface"]={}
            forceField["surface"]["type"]       = ["PatchyParticles","PatchyParticles"]

            forceField["surface"]["patchesState"]={}
            forceField["surface"]["patchesState"]["labels"]=["id","position"]
            forceField["surface"]["patchesState"]["data"]  =[]

            for i in range(self.nMonomers):
                index=i
                p = [0.0,self.monomerRadius*np.cos(self.beta0),self.monomerRadius*np.sin(self.beta0)]
                forceField["surface"]["patchesState"]["data"].append([int(index),list(p)])

            forceField["surface"]["patchesGlobal"]={}
            forceField["surface"]["patchesGlobal"]["fundamental"] = {}
            forceField["surface"]["patchesGlobal"]["fundamental"]["type"]      = ["Fundamental","None"]
            forceField["surface"]["patchesGlobal"]["fundamental"]["parameters"]= {}

            forceField["surface"]["patchesGlobal"]["types"]  = {}
            forceField["surface"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
            forceField["surface"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
            forceField["surface"]["patchesGlobal"]["types"]["data"]   = [["L",0.0,self.patchRadius,0.0]]

            forceField["surface"]["patchesTopology"] = {}

            forceField["surface"]["patchesTopology"]["structure"] = {}
            forceField["surface"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
            forceField["surface"]["patchesTopology"]["structure"]["data"]   = []

            for i in range(self.nMonomers):
                forceField["surface"]["patchesTopology"]["structure"]["data"].append([i,"L",i])

            forceField["surface"]["patchesTopology"]["forceField"] = {}

            forceField["surface"]["patchesTopology"]["forceField"]["linker"]={}
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["type"]       =  ["SurfacePatches", "Linker"]
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["parameters"] =  {"surfacePosition":self.plateBottom}
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["labels"]     =  ["name", "epsilon", "sigma"]
            forceField["surface"]["patchesTopology"]["forceField"]["linker"]["data"]       =  [["L",self.El,self.Sl]]

        ##############################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)

    def processSelection(self,**params):

        sel = []
        if "particleId" in params:
            sel += params["particleId"]
        return sel



import sys, os

import logging

from . import simulationStepBase

class savePatchyParticlesState(simulationStepBase):
    """
    Component name: savePatchyParticlesState
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 25/04/2023

    This component is used to save the state of the simulation incliding the
    patchy particles.

    Avalible formats are:
        * .coord
        * .sp
        * .xyz
        * .pdb
        * .itpv
        * .itpd
        * .dcd
        * .lammpstrj
        * .vel

    :param outputFilePath: Path to the output file
    :type outputFilePath: str
    :param outputFormat: Format of the output file
    :type outputFormat: str
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath","outputFormat","pbc"},
                         requiredParameters  = {"outputFilePath","outputFormat"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params.get("outputFilePath")
        parameters["outputFormat"]   = params.get("outputFormat")

        if "pbc" in params:
            parameters["pbc"] = params["pbc"]

        simulationStep = {
            name:{
              "type":["WriteStep","WritePatchyParticlesStep"],
              "parameters":{**parameters}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)






import logging

class BoundsBox:

    def __init__(self,nParticles,concentration):
        self.logger = logging.getLogger('VLMP')
        L = round((nParticles/concentration)**(1./3.),2)
        self.box = [L,L,L]

        self.logger.debug(f"[BoundsBox] Bounds box, box size : {L}")

    def getSimulationBox(self):
        return self.box

    def getSimulationBounds(self):
        return {}

class boxChecker:

    def __init__(self,box):

        self.box = box

    def check(self,position):

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        x,y,z = position
        if (x> boxX or y> boxY or z> boxZ or
            x<-boxX or y<-boxY or z<-boxZ):
            return False

        return True

class BoundsPlates:

    def __init__(self,nParticles,concentration,particleDiameter,padding,aspectRatio):
        self.logger = logging.getLogger('VLMP')

        self.particleDiameter = particleDiameter

        L = round((nParticles/(concentration*aspectRatio))**(1./3.),2)

        Lz = aspectRatio*L

        self.plateTop    =  Lz/2.0
        self.plateBottom = -Lz/2.0

        self.box = [L,L,self.plateTop + particleDiameter*padding - self.plateBottom]

        self.logger.debug(f"[BoundsPlates] Bounds box, box size : {L},{L},{Lz}")
        self.logger.debug(f"[BoundsPlates] Recomputed concentration {nParticles/(L*L*Lz)}")

    def getSimulationBox(self):
        return self.box

    def getSimulationBounds(self):
        return {"plateTop":self.plateTop,"plateBottom":self.plateBottom}

class platesChecker:

    def __init__(self,box,plateTop,plateBottom,particleDiameter):
        self.box              = box
        self.plateTop         = plateTop
        self.plateBottom      = plateBottom
        self.particleDiameter = particleDiameter

    def check(self,position):

        boxX,boxY,_ = [b/2.0 for b in self.box]

        x,y,z = position

        zSup    = self.plateTop    - 1.05*self.particleDiameter
        zBottom = self.plateBottom + 1.05*self.particleDiameter

        if (x> boxX or y> boxY or z>zSup or
            x<-boxX or y<-boxY or z<zBottom):
            return False

        return True





import sys, os

import logging

from . import typesBase

class basic(typesBase):

    """
    Component name: basic
    Component type: types

    Author: Pablo Ibáñez-Freire
    Date: 25/04/2023

    Basic types. Components:
    - mass: mass of the type
    - radius: radius of the type
    - charge: charge of the type

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = set(),
                         requiredParameters  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        self.setTypesName("Basic")

        self.addTypesComponent("mass", 1.0)
        self.addTypesComponent("radius", 1.0)
        self.addTypesComponent("charge", 0.0)



import os
import copy

import logging

################ MODEL INTERFACE ################

from pyUAMMD import simulation

from .. import idsHandler

from ...utils.selections import getSelections

class modelOperationBase(idsHandler):

    def __init__(self,
                 _type:str,_name:str,
                 units,types,ensemble,
                 models,
                 availableParameters:set,
                 requiredParameters:set,
                 availableSelections:set,
                 requiredSelections:set,
                 **params):
        super().__init__(models)

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self._units = units
        self._types = types
        self._ensemble = ensemble

        self._models = models

        self.logger.debug(f"[ModelOperation] ({self._type}) Operating on models: "+
                          " ".join([m.getName() for m in self._models])+
                          ". For model operation: "+self._name)

        self.availableParameters  = availableParameters.copy()
        self.requiredParameters   = requiredParameters.copy()
        self.availableSelections  = availableSelections.copy()
        self.requiredSelections   = requiredSelections.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[ModelOperation] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for model operation {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check all required selections are available selections
        if not self.requiredSelections.issubset(self.availableSelections):
            notAvailable = self.requiredSelections.difference(self.availableSelections)
            self.logger.error(f"[ModelOperation] ({self._type}) Some required selections ({notAvailable}) are not available selections for model operation {self._name}")
            raise Exception(f"Required selections are not available selections")

        # Check if all parameters and selectors given by params are available
        for par in params:
            if par not in self.availableParameters and par not in self.availableSelections:
                self.logger.error(f"[ModelOperation] ({self._type}) Parameter or selection {par} not available for model operation {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[ModelOperation] ({self._type}) Required parameter {par} not given for model operation {self._name}")
                raise Exception(f"Required parameter not given")

        # Check if all required selections are given
        for sel in self.requiredSelections:
            if sel not in params:
                self.logger.error(f"[ModelOperation] ({self._type}) Required selection {sel} not given for model operation {self._name}")
                raise Exception(f"Required selection not given")

        self.logger.info(f"[ModelOperation] ({self._type}) Using model operation {self._name}")

        ########################################################

        #Process selections
        selections = [sel for sel in params if sel in self.availableSelections]
        self._selection = getSelections(self._models,
                                        selections,
                                        **params)

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def getUnits(self):
        return self._units

    def getTypes(self):
        return self._types

    def getEnsemble(self):
        return self._ensemble

    def getModels(self):
        return self._models

    ########################################################

    def getSelection(self,selectionName):
        return self._selection[selectionName]

    ########################################################

    def getIdsProperty(self,ids,propertyName):
        return self._getIdsProperty(ids,propertyName)

    def getIdsState(self,ids,stateName):
        return self._getIdsState(ids,stateName)

    def getIdsStructure(self,ids,structName):
        return self._getIdsStructure(ids,structName)

    def setIdsState(self,ids,stateName,states):
        self._setIdsState(ids,stateName,states)


############### IMPORT ALL MODEL OPERATIONS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
operations = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
operations = [x.split(".")[0] for x in operations if "__" not in x]

for o in operations:
    try:
        exec(f"from .{o} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[ModelOperation] Error importing model operation type component {o}")



import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class interLennardJones(modelExtensionBase):

    """
    Component name: interLennardJones
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 21/09/2023

    Lennard-Jones interaction between all the particles in the system which belong to different molecules.

    :param interactionMatrix: Matrix of interaction parameters between different types of particles.
    :type interactionMatrix: list of lists. Each list is a row of the matrix. Format type1,type2,epsilon,sigma
    :param cutOffFactor: Factor to multiply the sigma parameter to obtain the cut-off distance.
    :type cutOffFactor: float
    :param addVerletList: If True, a Verlet list will be created for the interactions.
    :type addVerletList: bool, optional, default=True

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"interactionMatrix","cutOffFactor","addVerletList"},
                         requiredParameters  = {"interactionMatrix","cutOffFactor"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        cutOffFactor       = params.get("cutOffFactor")
        interactionMatrix  = params.get("interactionMatrix")

        addVerletList      = params.get("addVerletList",True)

        extension = {}

        if addVerletList:
            extension["nl"]={}
            extension["nl"]["type"]       =  ["VerletConditionalListSet",  "nonExclIntra_nonExclInter"]
            extension["nl"]["parameters"] =  {}
            extension["nl"]["labels"]      = ["id","id_list"]
            extension["nl"]["data"]        = []

        extension[name] = {}
        extension[name]["type"] = ["NonBonded","GeneralLennardJonesType2"]
        extension[name]["parameters"] = {"cutOffFactor":cutOffFactor,"condition":"inter"}
        extension[name]["labels"] = ["name_i","name_j","epsilon","sigma"]
        extension[name]["data"]   = interactionMatrix.copy()

        ############################################################

        self.setExtension(extension)






import logging
import copy

def getSelections(models,selectionsList,**param):

    logger = logging.getLogger("VLMP")

    #Check params
    for p in param.keys():
        if p not in selectionsList:
            continue
        else:
            for k in param[p].keys():
                if k not in ["models","expression"]:
                    logger.error(f"[getSelections] Parameter {k} not recognized for selection {p},"
                                  " available parameters are: models, expression")
                    raise Exception("Unknown parameter")

    selections = {}

    for sel in [s for s in param.keys() if s in selectionsList]:
        selections[sel] = []

        if "models" in param[sel]:
            selectedModels = [ m for m in models if m.getName() in param[sel]["models"]]
        else:
            selectedModels = models

        currentSelIdsLen = None
        for mdl in selectedModels:
            if "expression" in param[sel]:
                mdlSelIds = mdl.getSelection(**param[sel]["expression"])
            else:
                mdlSelIds = mdl.getSelection()

            offSet    = mdl.getIdOffset()

            #Check if mdlSelIds[0] is a list
            if isinstance(mdlSelIds[0],list):
                selIdsLen = len(mdlSelIds[0])
            else:
                selIdsLen = 1

            if currentSelIdsLen is None:
                currentSelIdsLen = selIdsLen
            elif currentSelIdsLen != selIdsLen:
                logger.error(f"[getSelections] Selection {sel} has different lengths")
                raise Exception("Selection has different lengths")

            for i in range(len(mdlSelIds)):
                if selIdsLen == 1:
                    mdlSelIds[i] += offSet
                else:
                    for j in range(selIdsLen):
                        mdlSelIds[i][j] += offSet

            selections[sel] += mdlSelIds

        if currentSelIdsLen == 1:
            selections[sel] = list(set(selections[sel]))
        else:
            selections[sel] = list(set([tuple(s) for s in selections[sel]]))

#            offSet    = mdl.getIdOffset()
#            mdlSelIds = [i+offSet for i in mdlSelIds]
#
#            selections[sel] += mdlSelIds
#        selections[sel] = list(set(selections[sel]))

    #Selection is a set of global ids
    return copy.deepcopy(selections)

def splitStateAccordingStructure(state,structure):

    logger = logging.getLogger("VLMP")

    #Check state and structure have the same length
    if len(state) != len(structure):
        logger.error("[splitStateAccordingStructure] State and structure have different lengths")
        raise Exception("State and structure have different lengths")

    # Structure has to have the following format:
    # [A,A,A,...,B,B,B,...,C,C,C,...,...]
    # Each letter represents a different structure
    # Check structure is correct
    appearedStructures = []
    for s in structure:
        if s not in appearedStructures:
            appearedStructures.append(s)
        elif s != appearedStructures[-1]:
            logger.error("[splitStateAccordingStructure] Structure is not correct")
            raise Exception("Structure is not correct")
        else:
            continue

    # Split pos according the different models
    splittedState = []

    currentStruct      = structure[0]
    currentStructState = []
    for i in range(len(structure)):
        if structure[i] == currentStruct:
            currentStructState.append(state[i])
        else:
            splittedState.append(currentStructState)
            currentStruct      = structure[i]
            currentStructState = [state[i]]

    splittedState.append(currentStructState)

    # At this point, splittedState is a list of lists,
    # each list contains the state of a model
    #                       struct[0]          struct[1]   ...
    # splittedState = [[state1_0,state2_0,...],[state1_0,state2_0,...],...]
    # It is ensured that the order is kept.
    # splittedState is the state list splitted according the structure

    return splittedState






import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class absortionSurface(modelExtensionBase):

    """
    Component name: absortionSurface
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 30/08/2023

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"heightThreshold","K"},
                         requiredParameters  = {"heightThreshold","K"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        heightThreshold = params["heightThreshold"]
        K               = params["K"]

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Surface","Absorbed"]
        extension[name]["parameters"] = {"heightThreshold":heightThreshold,"K":K}

        ############################################################

        self.setExtension(extension)



